/*
Java Graphics Demos: 
by Orion Lawlor, 12/1998
olawlor@acm.org

This code is totally free, and
should be considered in the public domain.
Use and modify it at will.
*/
import java.awt.*;
import java.awt.image.*;
import ColorMap;
import ByteTable;

class Demo 
{
	int w,h;
//Utilities
	static float dt=0.1f; //Timestep in seconds
	static float diffusionRate=10.0f; //Diffusion rate, pixels/sec
	static float diffusionBlob=dt*diffusionRate;
	static float heatLossRate=0.1f; //Heat loss rate, degrees/sec
	static float heatLossBlob=dt*heatLossRate;
	void blur(float pix[])
	{	//Blur the given image, doing nothing to boundary
	/*	for (int y=1;y<h-1;y++)
		{
			int l=y*w;
			for (int x=1;x<w-1;x++)*/
		
		//Blur all red pixels
		for (int y=1;y<h-1;y++)
		{
			int l=y*w;
			for (int x=1+(y%2);x<w-1;x+=2)
			{
				int x1=l+x-w;
				int x2=l+x;
				int x3=l+x+w;
				float nbor=
					pix[x1-1]+pix[x1]+pix[x1+1]+
					pix[x2-1]+        pix[x2+1]+
					pix[x3-1]+pix[x3]+pix[x3+1];
				pix[x2]+=diffusionBlob*(nbor*(1.0/8.0)-pix[x2])-heatLossRate;
			}
		}
		//Blur all black pixels
		for (int y=1;y<h-1;y++)
		{
			int l=y*w;
			for (int x=1+((y+1)%2);x<w-1;x+=2)
			{
				int x1=l+x-w;
				int x2=l+x;
				int x3=l+x+w;
				float nbor=
					pix[x1-1]+pix[x1]+pix[x1+1]+
					pix[x2-1]+        pix[x2+1]+
					pix[x3-1]+pix[x3]+pix[x3+1];
				pix[x2]+=diffusionBlob*(nbor*(1.0/8.0)-pix[x2])-heatLossRate;
			}
		}
	}

	float phase[];//Fluid phase (0-- solid; 1.0-- liquid) field
	float temperature[];//Fluid temperature (Degrees K) field
	byte outImage[];//Output image pixels
	static final float zeroCelcius=273.15f; //zero celcius in kelvin
	static final float freeze_pt=zeroCelcius;//Fluid freezes at this temperature (Degrees K)

	static final float loTemp=zeroCelcius-30.0f; //Temperature to display as black (Degrees K)
	static final float rangeTemp=80.0f; //Temperature range to display (Degrees K)
	static final float temp2byte=255.0f/rangeTemp;

//This much "temperature" is released during freezing process.
// (equal to heat of fusion / specific heat )
	static final float phase2temp=300.0f;
	static final float temp2phase=1.0f/phase2temp;

	//These constants indicate the direction the pixel was "colonized" from--
	// they give the local crystal orientation
	static final byte COL_TABLE[/* <V>*2+<H> */]= 
		{ (byte)240,//<- something is very wrong-- pixel is *not* colonized!
		  (byte)253,//Colonized in horizontal direction only
		  (byte)254,//Colonized in vertical direction only
		  (byte)255,//Colonized from both horiz. and vert. directions
		};
	public void init(ColorMap cm,byte pix[],int Nw,int Nh)
	{
		int x,y;
		w=Nw;h=Nh;
		int dex[]  ={0, 80,110,120,125,250, 253,254,255},
		    red[]  ={0,200,255,255,255,255,   0, 50,150},
		    green[]={0,  0,100,255,255,255,   0, 50,150},
		    blue[] ={0,  0, 50,  0,255,255, 150,250,150};
		cm.addBreaks(dex,red,green,blue);
		phase=new float[w*h];
		temperature=new float[w*h];
		outImage=pix;
		for (y=0;y<h;y++)
		{
			int l=y*w;
			for (x=0;x<w;x++) {
				phase[l+x]=1.0f;//Initially, all is liquid
				temperature[l+x]=(float)(freeze_pt-10.0+10.0f*2*(Math.random()-0.5f));
				outImage[l+x]=(byte)0;
			}
		}
		//Make a single solid (crystal seed) pixel
		x=w/2;y=h/2;
		phase[y*w+x]=0.0f;
		outImage[y*w+x]=(byte)255;
		//Spread the temerature a few times
		for (int i=0;i<3;i++)
			blur(temperature);
	}
	public boolean handleAction(Event ev,Object obj) {
		return false;
	}
	//Determines where this pixel is "colonized" from, and sets the
	// output array accordingly
	//Applies physical freeze modeling to given pixel index
	public void freezing(int dex)
	{
		float t=temperature[dex];
		float p=phase[dex];
		if (t<freeze_pt && p!=0)
		{//This pixel might be freezing-- check its neighbors
			boolean horiz=(phase[dex-1]==0.0f) || (phase[dex+1]==0.0f);
			boolean vert=(phase[dex-w]==0.0f) || (phase[dex+w]==0.0f);
			if (horiz||vert)
			{//This fluid can be frozen a bit-- it's near a frozen pixel
				if ((0xff&(int)outImage[dex])<240)
					outImage[dex]=COL_TABLE[(vert?2:0)+(horiz?1:0)];
				
				float dt=(freeze_pt-t);//Surrounding coldness we can up
				float dp=dt*temp2phase;//Phase change
				if (dp>p) dp=p;//<- keeps phase positive
				phase[dex]-=dp;
				//Deposit heat of fusion on neighbors
				float dt_final=dp*phase2temp;
				temperature[dex  ]+=dt_final;
			}
		}
	}
	int frameCount=0;
	public void fillBuffer(ColorMap cm,byte pix[])
	{
		//Physics pass(es)--goes in each direction to avoid bias
		int left=2,right=2,top=2,bot=2;
		for (int y=top;y<h-bot;y++)
		{
			int l=y*w; 
			for (int x=left;x<w-right;x++) 
				freezing(l+x);
		}
		
		//Propagate new temperatures
		blur(temperature);
		
		//Convert temperature for output
		for (int y=0;y<h;y++)
		{
			int l=y*w;
			for (int x=0;x<w;x++)
			{
				int dex=l+x;
				if ((0xff&(int)outImage[dex])<240) {
					double scaledTemp=(temperature[dex]-loTemp)*temp2byte;
					if (scaledTemp<0) outImage[dex]=0;
					else if (scaledTemp>240) outImage[dex]=(byte)240;
					else outImage[dex]=(byte)scaledTemp;
				}
			}
		}
	}
}
